import imblearn as _imblearn
import pandas as _pd


def resample(dataframe: _pd.DataFrame, label_column: str, sampling_strategy: str, k_neighbors = 5) -> _pd.DataFrame:
    """
    Splits the given DataFrame into a separate X / y DataFrame, then applies the SMOTE sampling
    strategy and finally merges the X / y DataFrame back to a new dataframe.

    'minority': resample only the minority class;
    'not minority': resample all classes but the minority class;
    'not majority': resample all classes but the majority class;
    'all': resample all classes;
    'auto': equivalent to 'not majority'.

    Args:
        dataframe: dataframe to use as input for oversampling.
        label_column: the column containing the label values.
        sampling_strategy: which sampling strategy should be applied
        k_neighbors: number of neighbors to use

    Returns:
        A new dataframe containing the original data and the oversampled data
    """
    y = dataframe[[label_column]]
    X = dataframe.drop(label_column, axis=1, inplace=False)
    x_s, y_s = resample_xy(X=X, y=y, sampling_strategy=sampling_strategy, k_neighbors=k_neighbors)
    return _pd.concat([x_s, y_s], axis=1)


def resample_xy(X: _pd.DataFrame, y: _pd.DataFrame, sampling_strategy: str, k_neighbors=5) -> (_pd.DataFrame, _pd.DataFrame):
    """
    Applies the SMOTE sampling strategy and finally merges the X / y DataFrame back to a new dataframe.

    'minority': resample only the minority class;
    'not minority': resample all classes but the minority class;
    'not majority': resample all classes but the majority class;
    'all': resample all classes;
    'auto': equivalent to 'not majority'.

    Args:
        X: DataFrame containing the features
        y: DataFrame containing the labels
        sampling_strategy: which sampling strategy should be applied
        k_neighbors: number of neighbors to use

    Returns:
        A new dataframe containing the original data and the oversampled data
    """
    smote = _imblearn.over_sampling.SMOTE(sampling_strategy=sampling_strategy, k_neighbors=k_neighbors)
    return smote.fit_resample(X=X, y=y)


import numpy as _np
import tensorflow as _tf
import re as _re
from typing import Union as _Union

########################################################################################################################
# Word level functions
########################################################################################################################

# Based on TextVectorization
PUNCTUATION = r'[!"#$%&()\*\+,-\./:;<=>?@\[\\\]^_`{|}~\']'
LOWER_AND_STRIP_PUNCTUATION = "lower_and_strip_punctuation"
STRIP_PUNCTUATION = "strip_punctuation"
LOWER = "lower"


def count_unique_words(lines: _Union[list[str], _tf.Tensor], standardize="lower_and_strip_punctuation") -> int:
    if _tf.is_tensor(x=lines):
        lines = lines.numpy()

    regex = _re.compile(PUNCTUATION)

    unique = set()
    for line in lines:
        if isinstance(line, (bytes, bytearray)):
            line = line.decode(encoding='utf-8')

        if standardize in (LOWER, LOWER_AND_STRIP_PUNCTUATION):
            line = line.lower()
        if standardize in (STRIP_PUNCTUATION, LOWER_AND_STRIP_PUNCTUATION):
            line = regex.sub('', line)
        unique = unique.union(set(line.split()))
    return len(unique)


def calculate_q_precentile_word_lengths(lines: _Union[list[str], _tf.Tensor], q=95) -> int:
    """
    Calculates the q-percentile based on the lengths of the strings.

    :param lines: a list of string values
    :param q: q-percentile (default 95%)
    :return: the percentile
    """
    if _tf.is_tensor(x=lines):
        lines = lines.numpy()

    return int(_np.percentile(a=__count_words_for_each_sentence(lines), q=q))


def maximum_number_of_word_per_sentence(lines: _Union[list[str], _tf.Tensor]) -> int:
    """
    Returns the maximum number of words per sentence.


    :param lines: a list of string values
    :return: the max sentence length in words
    """
    if _tf.is_tensor(x=lines):
        lines = lines.numpy()

    return max(__count_words_for_each_sentence(lines))


def average_number_of_word_per_sentence(lines: _Union[list[str], _tf.Tensor]) -> int:
    """
    Returns the average number of words per sentence.

    :param lines: a list of string values
    :return: the average sentence length in words
    """
    if _tf.is_tensor(x=lines):
        lines = lines.numpy()

    return round(sum(__count_words_for_each_sentence(lines)) / len(lines))


########################################################################################################################
# Character level functions
########################################################################################################################

def count_unique_chars(lines: _Union[list[str], _tf.Tensor], standardize="lower_and_strip_punctuation") -> (int, list[chr]):
    regex = _re.compile(PUNCTUATION)

    if _tf.is_tensor(x=lines):
        lines = lines.numpy()

    unique = set()
    for line in lines:
        if isinstance(line, (bytes, bytearray)):
            line = line.decode(encoding='utf-8')

        if standardize in (LOWER, LOWER_AND_STRIP_PUNCTUATION):
            line = line.lower()
        if standardize in (STRIP_PUNCTUATION, LOWER_AND_STRIP_PUNCTUATION):
            line = regex.sub('', line)
        unique = unique.union(set(list(line)))
    return len(unique), unique


def length_per_sentence(lines: _Union[list[str], _tf.Tensor]) -> list[int]:
    """
    Calculates the length for each string value and returns a list containing
    these lengths.

    :param lines: a list of string values
    :return:
    """
    if _tf.is_tensor(x=lines):
        lines = lines.numpy()

    return [len(sentence) for sentence in lines]


def calculate_q_precentile_character_lengths_for_all_sentence(lines: _Union[list[str], _tf.Tensor], q=95) -> int:
    """
    Calculates the q-percentile based on the lengths of the strings.

    :param lines: a list of string values
    :param q: q-percentile (default 95%)
    :return: the percentile
    """
    return int(_np.percentile(a=length_per_sentence(lines), q=q))


def calculate_average_character_length_for_all_sentences(lines: _Union[list[str], _tf.Tensor]) -> int:
    """
    Returns the average length of the strings.

    :param lines: a list of string values
    :return: the average sentence length
    """
    if _tf.is_tensor(x=lines):
        lines = lines.numpy()

    return round(sum(length_per_sentence(lines)) / len(lines))


def __count_words_for_each_sentence(lines: _Union[list[str], _tf.Tensor]) -> list[int]:
    """
    Calculates the length for each string value and returns a list containing
    these lengths.

    :param lines: a list of string values
    :return:
    """
    if _tf.is_tensor(x=lines):
        lines = lines.numpy()

    return [len(i.split()) for i in lines]



import tensorflow as _tf
import numpy as _np


def make_naive_predictions(y_true):
    """
    Method that creates a naive prediction from the given truth labels:

    Input:
        [0, 1, 2, 3, 4, 5, 6, 7]

    Output:
        [1, 2, 3, 4, 5, 6, 7]
        [0, 1, 2, 3, 4, 5, 6]


    Args:
        y_true: the truth labels for which we want to make a naive prediction

    Returns:
        y_true and the naive prediction based on the given truth labels

    """
    naive_predictions = y_true[:-1]
    return y_true[1:], naive_predictions


def make_windows(x, window_size, horizon):
    """
    Turns 1 1D array into a 2D array of sequential labelled windows of window_size with horizon size labels.

    E.g. if horizon = 1 and window_size = 3, then
    X = [1, 2, 3, 4, 5, 6, 7, 8, 9]

    Output:
        Features:
        [
            [0, 1, 2, 3, 4, 5, 6]
            [1, 2, 3, 4, 5, 6, 7]
        ]

        Labels:
        [
            [7]
            [8]
        ]

    Args:
        x: A 1D Numpy array or Tensor or a single column Pandas dataframe
        window_size: the number of time steps to use as features
        horizon:  the number of time steps to use as labels

    """
    if not isinstance(x, _tf.Tensor):
        x = _tf.convert_to_tensor(value=x)

    # Check that we only have a single feature
    assert x.ndim == 1, (f"Shape of x is not correct {x.shape}, should be 1D array (ie. [1, 2, 3, 4]), please check "
                         f"the following:"
                         f"* Does your data contain multiple columns, if so, select only a single column, if it is a "
                         f"Pandas dataframe then use df['col'], if it is a tensor use slicing, for example t[:,2] to "
                         f"select the third column, If your data is column based then convert it to row based "
                         f"using tf.squeeze()")

    x = x.numpy()

    # 1. Create a window of specific window_size (add the horizon on the end for labelling later)
    window_step = _np.expand_dims(_np.arange(window_size + horizon), axis=0)

    # 2. Create a 2D array for multiple window step (minus 1 to account for 0 indexing)
    window_indexes = window_step + _np.expand_dims(_np.arange(len(x) - (window_size + horizon - 1)),
                                                  axis=0).T  # Create 2D array of windows of window size window_size.

    # 3. Index on the target array (a time series) with 2D array of multiple window steps
    windowed_array = x[window_indexes]

    # 4. get the labelled windows
    windows, labels = _get_labelled_windows(windowed_array, horizon=horizon)
    return windows, labels


def _get_labelled_windows(x, horizon):
    """
    Creates labels for windowed dataset.

    E.g. if horizon = 1
    Input: [0, 1, 2, 3, 4, 5, 6, 7] -> Output: ([0, 1, 2, 3, 4, 5, 6], [7])
    """
    return x[:, :-horizon], x[:, -horizon:]


from tensorflow.python.platform import tf_logging as _logging
from keras.callbacks import Callback as _Callback
from keras.callbacks import TensorBoard as _TensorBoard
from keras.callbacks import LearningRateScheduler as _LearningRateScheduler
from keras.callbacks import ModelCheckpoint as _ModelCheckpoint
from keras.callbacks import EarlyStopping as _EarlyStopping
from keras.callbacks import ReduceLROnPlateau as _ReduceLROnPlateau

import datetime as _datetime
import os as _os


def reduce_lr_on_plateau_callback(monitor="val_loss",
                                  factor=0.7,
                                  patience=4,
                                  verbose=1,
                                  mode="auto",
                                  min_delta=1e-4,
                                  cooldown=0,
                                  min_lr=1e-7) -> _ReduceLROnPlateau:
    return _ReduceLROnPlateau(monitor=monitor,
                              factor=factor,
                              patience=patience,
                              verbose=verbose,
                              mode=mode,
                              min_delta=min_delta,
                              cooldown=cooldown,
                              min_lr=min_lr)


def early_stopping_callback(monitor='val_loss',
                            min_delta=0,
                            patience=6,
                            verbose=1,
                            mode='auto',
                            baseline=None,
                            restore_best_weights=True,
                            start_from_epoch=0) -> _EarlyStopping:
    return _EarlyStopping(monitor=monitor,
                          min_delta=min_delta,
                          patience=patience,
                          verbose=verbose,
                          mode=mode,
                          baseline=baseline,
                          restore_best_weights=restore_best_weights,
                          start_from_epoch=start_from_epoch)


def model_checkpoint_callback(filepath='/path/to/model.h5',
                              monitor='val_loss',
                              save_weights_only=True,
                              save_best_only=True,
                              save_freq='epoch',
                              verbose=0) -> _ModelCheckpoint:

    return _ModelCheckpoint(filepath=filepath,
                            monitor=monitor,
                            save_weights_only=save_weights_only,
                            save_best_only=save_best_only,
                            save_freq=save_freq,
                            verbose=verbose)


def learning_rate_scheduler_callback(learning_rate_start=0.001, epochs=50) -> _LearningRateScheduler:
    """
    Creates a LearningRateScheduler which will be pre-configured with a division. The division
    is calculated using find_learning_rate_division. At the start of every epoch a new learning rate
    will be configured.

    :param learning_rate_start: initial starting learning rate
    :param epochs: number of epochs the model will train for
    :return: the pre-configured LearningRateScheduler
    """
    min, max, division = _find_learning_rate_division(learning_rate=learning_rate_start, epochs=epochs)
    print(f"Min learning rate: {min}\nMax learning rate: {max}\nDivision: {division}")
    return _LearningRateScheduler(lambda epoch: learning_rate_start * 10 ** (epoch / division))


def tensorboard_callback(experiment_name: str, dir_name='./logs'):
    log_dir = dir_name + '/' + experiment_name + '/' + _datetime.datetime.now().strftime('%Y%m%d-%H%M%S')
    print(f"Saving TensorBoard log files to: {log_dir}")
    return _TensorBoard(log_dir=log_dir)


def create_stop_training_on_target_callback(metric="val_loss",
                                            target=0.95,
                                            patience=0,
                                            verbose=0,
                                            start_from_epoch=0,
                                            restore_best_weights=False):
    return StopTrainingOnTarget(metric=metric,
                                target=target,
                                patience=patience,
                                verbose=verbose,
                                start_from_epoch=start_from_epoch,
                                restore_best_weights=restore_best_weights)


def _find_learning_rate_division(learning_rate: float, epochs: int):
    """
    Finds the optimal division which can be used in a learning rate scheduler.

    epochs = 50
    initial_lr = 0.001

    division = find_learning_rate_division(initial_lr, epochs)
    lr_scheduler = LearningRateScheduler(lambda epoch: initial_lr * 10 ** (epoch / division))

    :param learning_rate: initial learning rate
    :param epochs: number of epochs that the model will train for
    :return: the minimum learning rate, maximum learning rate and the division which can be used in the scheduler.
    """
    min_lr = 0.
    max_lr = 0.
    division = 1000
    while max_lr < 0.1:
        min_lr = learning_rate * 10 ** (1 / division)
        max_lr = learning_rate * 10 ** (epochs / division)
        division -= 1
    return min_lr, max_lr, division


class StopTrainingOnTarget(_Callback):

    def __init__(self,
                 metric="val_loss",
                 target=0.95,
                 patience=0,
                 verbose=0,
                 start_from_epoch=0,
                 restore_best_weights=False
                 ):
        super().__init__()
        self.metric = metric
        self.target = target
        self.patience = patience
        self.verbose = verbose
        self.start_from_epoch = start_from_epoch
        self.restore_best_weights = restore_best_weights
        self.stopped_epoch = 0
        self.wait = 0
        self.best_weights = None
        self.best_epoch = 0
        self.best_value = 0

    def on_train_begin(self, logs=None):
        # Allow instances to be re-used
        self.stopped_epoch = 0
        self.wait = 0
        self.best_weights = None
        self.best_epoch = 0
        self.best_value = 0

    def on_epoch_end(self, epoch, logs=None):
        current_metric_value = self._get_metric_value(logs)
        if current_metric_value is None or epoch < self.start_from_epoch:
            # If the metric is not part of the logs or still in initial warm-up stage.
            return

        self.wait += 1
        if self._is_metric_reached_or_better(current_metric_value):
            self.best_value = current_metric_value
            self.best_epoch = epoch
            self.wait = 0

        # If our patience has run out, and we have a best performing epoch
        if self.wait >= self.patience and self.best_epoch > 0:
            self.stopped_epoch = epoch
            self.model.stop_training = True
            if self.restore_best_weights and self.best_weights is not None:
                if self.verbose > 0:
                    print(f'Restoring model weights from the end of the best epoch: {self.best_epoch +1})')
                self.model.set_weights(self.best_weights)

    def on_train_end(self, logs=None):
        if self.stopped_epoch > 0 and self.verbose > 0:
            print(f"Epoch {self.stopped_epoch + 1}: training stopped "
                f"on metric a`{self.metric}` with a target value of {self.target}), "
                f"epoch {self.best_epoch} achieved {round(self.best_value * 100, 2)}% accuracy")

    def _get_metric_value(self, logs):
        logs = logs or {}
        value = logs.get(self.metric)
        if value is None:
            _logging.warning(
                f"Stop training goal condition on metric `{self.target}` "
                f"which is not available. Available metrics are: {','.join(list(logs.keys()))}"
            )
        return value

    def _is_metric_reached_or_better(self, value):
        # If the value is larger than the target and larger than the current known best value
        return self.target < value > self.best_value



import tensorflow as _tf
import numpy as _np
import pandas as _pd


def mean_absolute_scaled_error(y_true, y_pred, seasonality: int = None):
    """
    Calculates the mean absolute scaled error

    Args:
        y_true: the ground truth
        y_pred: the predictions
        seasonality: (int) the seasonality, 7=weekly, etc...

    Returns:
        The mean absolute scaled error
    """
    y_true = _to_numpy(y_true)
    y_pred = _to_numpy(y_pred)

    mae = _np.mean(_np.abs(y_true - y_pred))

    # Find MAE of naive forecast (no seasonality)
    if seasonality is not None:
        mad = _np.mean(_np.abs(y_true[seasonality:] - y_true[:-seasonality]))
    else:
        mad = _np.mean(_np.abs(y_true[1:] - y_true[:-1]))

    return mae / mad


def _to_numpy(x):
    if not isinstance(x, _np.ndarray):
        if isinstance(x, _pd.DataFrame) | isinstance(x, _pd.Series) | isinstance(x, _pd.DatetimeIndex):
            return x.to_numpy()
        elif isinstance(x, _tf.Tensor):
            return x.numpy()
        else:
            return _tf.convert_to_tensor(value=x).numpy()

    return x


import tensorflow as _tf
from typing import Union


def collect_layers(model: Union[_tf.keras.Model, _tf.keras.Sequential], recursive=True, include_trainable=True, include_non_trainable=True) -> list:
    """
    Collects layers from the given model, if the model is functional then there might be sub-branches
    inside the model, for example:

    data_augmentation = Sequential([
        RandomFlip("horizontal"),
        RandomHeight(.2),
        RandomWidth(.2),
        RandomRotation(.2),
        RandomZoom(.2),
    ], name="data_augemtation")

    inputs = Input(shape=INPUT_SHAPE, name="input_layer")
    x = data_augmentation(inputs)
    x = base_model(x, training=False)
    x = GlobalAveragePooling2D(name='global_average_pooling_2d')(x)
    outputs = Dense(10, activation="softmax", name="output_layer")(x)
    model_2 = Model(inputs, outputs)

    Where in this case the data_augmentation is a sub-branch inside the functional model, use recursive to also
    traverse these layers.

    :param model: the model
    :param recursive: follow recursive
    :param include_trainable: include layers that are set as trainable (default: True)
    :param include_non_trainable: include layers that are set as non-trainable (default: True)
    :return: the layers
    """
    layers = []

    for layer in model.layers:
        if hasattr(layer, 'layers') and recursive:
            layers = [*layers, *collect_layers(layer, recursive=recursive, include_trainable=include_trainable,
                                               include_non_trainable=include_non_trainable)]
        else:
            if (include_trainable and layer.trainable) or (include_non_trainable and layer.trainable is False):
                layers.append(layer)

    return layers


def collect_layer_names(model: Union[_tf.keras.Model, _tf.keras.Sequential], recursive=True, include_trainable=True, include_non_trainable=True) -> \
        list[str]:
    layers = collect_layers(model, recursive=recursive, include_trainable=include_trainable,
                            include_non_trainable=include_non_trainable)
    return list(map(lambda layer: layer.name, layers))


def list_model(model: Union[_tf.keras.Model, _tf.keras.Sequential], recursive=True, include_trainable=True, include_non_trainable=True) -> None:
    layers = collect_layers(model, recursive=recursive, include_trainable=include_trainable,
                            include_non_trainable=include_non_trainable)

    assert len(layers) > 0, f"Model has no layers."

    list_layers(layers=layers, include_trainable=include_trainable, include_non_trainable=include_non_trainable)

    total_params_s, trainable_params_s, non_trainable_params_s = calculate_parameters(object=model)
    print(f"\nTotal params: {total_params_s}")
    print(f"Trainable params: {trainable_params_s}")
    print(f"Non-trainable params: {non_trainable_params_s}")


def list_layers(layers: list[_tf.keras.layers.Layer], include_trainable=True, include_non_trainable=True) -> None:
    assert len(layers) > 0, f"List of layers is empty."

    layer_name_col_width = len(max(list(map(lambda l: l.name, layers)), key=len))
    layer_type_col_width = len(max(list(map(lambda l: type(l).__name__, layers)), key=len))

    # Filter to have only called layers.
    called_layers = filter(lambda layer: len(layer._inbound_nodes) > 0, layers)
    layer_shape_col_width = len(
        max(list(map(lambda l: str(l.output_shape), called_layers)), key=len, default='output_shape'))

    layer_dtype_col_width = len(max(list(map(lambda l: str(l.dtype), layers)), key=len))
    layer_dtype_policy_col_width = len(max(list(map(lambda l: str(l.dtype_policy.name), layers)), key=len))

    print(
        f"{'row':<5} | "
        f"{'name (type)':<{layer_name_col_width + layer_type_col_width + 3}} | "
        f"{'dtype':<{layer_dtype_col_width}} | "
        f"{'policy':<{layer_dtype_policy_col_width}} | "
        f"trainable | "
        f"{'output shape':<{layer_shape_col_width}} | "
        f"{'Total Param #'} | "
        f"{'Trainable Param #'} | "
        f"{'Non-trainable Param #'}")
    for layer_number, layer in enumerate(layers):
        if (include_trainable and layer.trainable) or (include_non_trainable and layer.trainable is False):
            total_params_s, trainable_params_s, non_trainable_params_s = calculate_parameters(object=layer)
            output_shape = 'unknown' if not layer.built else layer.output_shape
            print(
                f"{layer_number:<5} | "
                f"{layer.name:<{layer_name_col_width}} ({type(layer).__name__:<{layer_type_col_width}}) | "
                f"{str(layer.dtype):<{layer_dtype_col_width}} | "
                f"{str(layer.dtype_policy.name):<{layer_dtype_policy_col_width}} | "
                f"{str(layer.trainable):<9} | "
                f"{str(output_shape):<{layer_shape_col_width}} | "
                f"{total_params_s:<{len('Total Param #')}} | "
                f"{trainable_params_s:<{len('Trainable Param #')}} | "
                f"{non_trainable_params_s:<{len('Non-trainable Param #')}}")


def calculate_parameters(object: Union[_tf.keras.Model, _tf.keras.Sequential, _tf.keras.layers.Layer]) -> (str, str, str):
    total_params_s = trainable_params_s = non_trainable_params_s = 'unknown'
    if object.built:
        all_layers = collect_layers(object, recursive=True, include_trainable=True, include_non_trainable=True)

        total_params = sum([_tf.size(var).numpy() for layer in all_layers for var in layer.variables])
        trainable_params = sum([_tf.size(var).numpy() for layer in all_layers for var in layer.trainable_variables])
        non_trainable_params = total_params - trainable_params

        total_params_s = f"{total_params:,}"
        trainable_params_s = f"{trainable_params:,}"
        non_trainable_params_s = f"{non_trainable_params:,}"

    return total_params_s, trainable_params_s, non_trainable_params_s



import tensorflow as _tf
import numpy as _np


def set_seed(seed):
    """
    Set the seed value for TensorFlow, Keras and Numpy.
    Args:
        seed:

    Returns:

    """
    _tf.random.set_seed(seed=seed)
    _tf.keras.utils.set_random_seed(seed=seed)
    _np.random.seed(seed=seed)
